package com.gxzc.zen.common.util

import com.github.benmanes.caffeine.cache.Caffeine
import com.gxzc.zen.common.dto.ZenFileMetadata
import com.gxzc.zen.common.dto.ZenFileResponse
import com.gxzc.zen.common.exception.ZenException
import com.gxzc.zen.common.exception.ZenExceptionEnum
import com.gxzc.zen.common.properties.UploadProperties
import org.apache.commons.io.FilenameUtils
import org.slf4j.LoggerFactory
import org.springframework.cache.caffeine.CaffeineCache
import org.springframework.web.multipart.MultipartFile
import java.io.*
import java.nio.file.Files
import java.nio.file.Paths
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.TimeUnit

/**
 * 上传 工具类
 * @author NorthLan
 * @date 2018/5/19
 * @url https://noahlan.com
 */
object UploadUtil {
    private val logger = LoggerFactory.getLogger(UploadUtil::class.java)

    private var uploadProperties = SpringContextHolder.getBean(UploadProperties::class.java)
        get() {
            if (field == null) {
                field = SpringContextHolder.getBean(UploadProperties::class.java)
            }
            return field
        }

    private val uploadedStatusMap = ConcurrentHashMap<String, String>() // 文件上传状态缓存
    private val batchCountMap = CaffeineCache("batchCountMap", Caffeine.newBuilder().expireAfterWrite(1L, TimeUnit.HOURS).build(), true) // 文件批量上传批次文件数量记录

    private val FILE_SEPARATOR = System.getProperty("file.separator") // 适配操作系统的文件路径

    object STATUS {
        const val CHECKING = "checking" // 检查中
        const val MERGING = "merging" // 合并中
        const val UPLOADED = "uploaded" // 单文件上传完毕
        const val BATCH_UPLOADED = "batchUploaded" // 单文件上传完毕
        const val UPLOADING = "uploading" // 上传中
    }

    /**
     * 正常上传<br>
     * 文件夹结构/文件名
     */
    fun upload(fileMetadata: ZenFileMetadata, file: MultipartFile): ZenFileResponse {
        val tmpPath = uploadProperties!!.tmpPath!!
        val dataPath = uploadProperties!!.dataPath!!
        val chunkSize = uploadProperties!!.chunkSize!!
        var retFile: File? = null
        if (validateRequest(fileMetadata, file)) {
            // 文件分片 如果分片小于chunkSize && totalChunk = 1,直接转存
            if (fileMetadata.totalChunks == 1 && fileMetadata.chunkSize!! <= chunkSize) {
                val filename = fileMetadata.filename!!
                val relativePath = fileMetadata.relativePath!!

                val destPath = FilenameUtils.normalize("$dataPath$FILE_SEPARATOR${getDestFilePath(filename, relativePath, fileMetadata.repath)}")
                val destDir = Paths.get(destPath)
                if (Files.notExists(destDir)) {
                    Files.createDirectories(destDir)
                }
                val outputPath = FilenameUtils.normalize("$destPath$FILE_SEPARATOR${getDestFileName(filename, fileMetadata.rename)}")
                retFile = File(outputPath)
                file.transferTo(retFile)
                retFile.setLastModified(fileMetadata.lastModified!!)
            } else {
                val chunkFilename = getChunkFilename(tmpPath, fileMetadata.chunkNumber, fileMetadata.md5)
                val directory = Paths.get(tmpPath)
                if (Files.notExists(directory)) {
                    Files.createDirectories(directory)
                }
                file.transferTo(File(chunkFilename))
                // 检查分块完整性
                if (checkChunks(tmpPath, fileMetadata)) {
                    // 合并
                    retFile = try {
                        mergeChunks(tmpPath, dataPath, fileMetadata)
                    } catch (e: Throwable) {
                        logger.error("merge file chunks exception, cause ", e)
                        uploadedStatusMap.remove(fileMetadata.md5!!)
                        null
                    }
                }
            }
        } else {
            throw ZenException(ZenExceptionEnum.FILE_METADATA_VALIDATE_ERROR)
        }
        var status: String = STATUS.UPLOADING
        if (retFile != null) {
            status = setBatch(fileMetadata)
        }
        return ZenFileResponse().apply {
            this.status = status
            this.file = retFile
        }
    }

    private fun setBatch(fileMetadata: ZenFileMetadata): String {
        val batchId = fileMetadata.batchId!!
        val totalNumber = fileMetadata.totalNumber!!
        synchronized(batchCountMap) {
            var count = batchCountMap[batchId]?.get() as? Int ?: 0
            return if (++count >= totalNumber) {
                batchCountMap.evict(batchId)
                STATUS.BATCH_UPLOADED
            } else {
                batchCountMap.put(batchId, count)
                STATUS.UPLOADED
            }
        }
    }

    /**
     * 检测文件是否存在 实现文件秒传
     */
    fun checkUpload(fileMetadata: ZenFileMetadata): ZenFileResponse {
        val dataPath = uploadProperties!!.dataPath!!
        val ret = ZenFileResponse()
        if (validateRequest(fileMetadata, null)) {
            if (fileExists(fileMetadata)) {
                ret.uploadedChunks = mutableListOf()
                for (i in 1..fileMetadata.totalChunks!!) {
                    ret.uploadedChunks?.add(i)
                }
                ret.status = setBatch(fileMetadata)
                ret.file = File(getFullDestFilename(dataPath, fileMetadata.filename!!, fileMetadata.relativePath!!, fileMetadata.repath, fileMetadata.rename))
            } else {
                // 检查分片存在情况
                for (i in 1..fileMetadata.totalChunks!!) {
                    if (chunkExists(i, fileMetadata.md5!!)) {
                        if (ret.uploadedChunks == null) {
                            ret.uploadedChunks = mutableListOf()
                        }
                        ret.uploadedChunks?.add(i)
                    }
                }
                if (ret.uploadedChunks != null && ret.uploadedChunks!!.size == fileMetadata.totalChunks) {
                    ret.status = STATUS.CHECKING // 秒传
                    // 所有分块传完 移除其中一个分块,再传一次而后 merge
                    ret.uploadedChunks!!.removeAt(0)
                }
            }
        }
        return ret
    }

    /**
     * 检查文件是否存在 (真实文件)
     */
    private fun fileExists(fileMetadata: ZenFileMetadata): Boolean {
        val dataPath = uploadProperties!!.dataPath!!
        // 获取filename 真实文件
        val filename = fileMetadata.filename!!
        val relativePath = fileMetadata.relativePath!!
        val md5 = fileMetadata.md5!!
        val path = getFullDestFilename(dataPath, filename, relativePath, fileMetadata.repath, fileMetadata.rename)
        return Files.exists(Paths.get(path)) && FileUtil.md5HeadTail(path, uploadProperties!!.chunkSize!!.toInt()) == md5 // # 防篡改
    }

    /**
     * 检查分片是否存在 (tmp目录)
     */
    private fun chunkExists(chunkNumber: Int?, md5: String): Boolean {
        val tmpPath = uploadProperties!!.tmpPath!!
        return Files.exists(Paths.get(getChunkFilename(tmpPath, chunkNumber, md5)))
    }

    /**
     * 检查分块是否上传完毕
     */
    private fun checkChunks(tmpPath: String, fileMetadata: ZenFileMetadata): Boolean {
        val totalChunks = fileMetadata.totalChunks!!
        val md5 = fileMetadata.md5!!
        for (i in 1..totalChunks) {
            if (Files.notExists(Paths.get(getChunkFilename(tmpPath, i, md5)))) {
                return false
            }
        }
        return true
    }

    /**
     * 合并所有分块
     * 支持 文件夹合并
     */
    private fun mergeChunks(sourceRootPath: String, destRootPath: String, fileMetadata: ZenFileMetadata): File? {
        // 源文件的文件夹信息 然后组合拼接
        val filename = fileMetadata.filename!!
        val relativePath = fileMetadata.relativePath!!
        val totalChunks = fileMetadata.totalChunks!!
        val md5 = fileMetadata.md5!!

        if (uploadedStatusMap[md5] == STATUS.MERGING) {
            return null
        }

        uploadedStatusMap[md5] = STATUS.MERGING

        val t = System.currentTimeMillis()
        logger.debug("start merging chunks for [$filename], now: $t")

        val destPath = getFullDestFilepath(destRootPath, filename, relativePath, fileMetadata.repath, fileMetadata.rename)
        val destDir = Paths.get(destPath)
        if (Files.notExists(destDir)) {
            Files.createDirectories(destDir)
        }
        // 目标文件流 通过文件名来拼接
        val outputPath = getFullDestFilename(destRootPath, filename, relativePath, fileMetadata.repath, fileMetadata.rename)
        val outputFile = File(outputPath)
        val destOutputStream = BufferedOutputStream(FileOutputStream(outputFile))

        val buffer = ByteArray(4096)
        var readBytesLength: Int
        for (i: Int in 1..totalChunks) {
            val sourceFile = File(getChunkFilename(sourceRootPath, i, md5))
            val sourceInputStream = BufferedInputStream(FileInputStream(sourceFile))

            readBytesLength = sourceInputStream.read(buffer)
            while (readBytesLength != -1) {
                destOutputStream.write(buffer, 0, readBytesLength)
                readBytesLength = sourceInputStream.read(buffer)
            }
            sourceInputStream.close()
            // 删除分片
            sourceFile.delete()
        }
        //
        destOutputStream.flush()
        destOutputStream.close()


        // 修改文件 修改时间
        outputFile.setLastModified(fileMetadata.lastModified!!)

        uploadedStatusMap.remove(md5)

        logger.debug("merging successful, cost: ${System.currentTimeMillis() - t} ms.")
        return outputFile
    }

    /**
     * 获取 分片文件名
     */
    private fun getChunkFilename(path: String, chunkNumber: Int?, identifier: String?): String {
        return "$path/upload-$identifier.$chunkNumber"
    }

    /**
     * 生成 输出的文件名
     */
    private fun getDestFileName(filename: String, rename: String?): String {
        return if (rename == null || rename.isEmpty()) {
            filename
        } else {
            // 后缀
            val ext = FilenameUtils.getExtension(rename)
            if (ext.isNullOrEmpty()) {
                "$rename.${FilenameUtils.getExtension(filename)}"
            } else {
                rename
            }
        }
    }

    /**
     * 生成 输出的文件夹结构
     */
    private fun getDestFilePath(filename: String, relativePath: String, repath: String?): String {
        return if (repath.isNullOrEmpty()) {
            relativePath.replace(filename, "")
        } else {
            repath!!
        }
    }

    /**
     * 获取输出完整文件路径
     */
    private fun getFullDestFilepath(dataPath: String, filename: String, relativePath: String, repath: String?, rename: String?): String {
        return FilenameUtils.getFullPath(getFullDestFilename(dataPath, filename, relativePath, repath, rename))
    }

    /**
     * 获取输出完整文件名
     */
    private fun getFullDestFilename(dataPath: String, filename: String, relativePath: String, repath: String?, rename: String?): String {
        return FilenameUtils.normalize("$dataPath$FILE_SEPARATOR${getDestFilePath(filename, relativePath, repath)}$FILE_SEPARATOR${getDestFileName(filename, rename)}")
    }

    /**
     * validate request multipart chunks
     */
    private fun validateRequest(fileMetadata: ZenFileMetadata, file: MultipartFile?): Boolean {
        val md5 = fileMetadata.md5
        val chunkNumber = fileMetadata.chunkNumber
        val chunkSize = fileMetadata.chunkSize
        val totalSize = fileMetadata.totalSize
        val filename = fileMetadata.filename

        if (chunkNumber == null || chunkNumber <= 0 ||
                chunkSize == null || chunkSize <= 0 ||
                totalSize == null || totalSize <= 0 ||
                md5 == null || md5.isEmpty() ||
                filename == null || filename.isEmpty()) {
            return false
        }

        val numberOfChunks = Math.max(Math.floor(totalSize / (chunkSize * 1.0)), 1.0).toInt()
        if (chunkNumber > numberOfChunks) {
            return false
        }

        // is the file too large?
        uploadProperties?.let {
            val maxFileSize = it.maxFileSize
            if (maxFileSize != null && maxFileSize > 0) {
                if (totalSize > maxFileSize) {
                    logger.error("filesize limit: [${maxFileSize / 1024 / 1024} MB], now [${totalSize / 1024 / 1024} MB]")
                    return false
                }
            }
        }

        if (file != null) {
            // The chunk in the POST request isn't the correct size
            if (chunkNumber < numberOfChunks && file.size != chunkSize) {
                logger.error("The chunk in the POST request isn't the correct size")
                return false
            }
            // The chunks in the POST is the last one, and the fil is not the correct size
            if (numberOfChunks > 1 && chunkNumber == numberOfChunks && file.size != (totalSize % chunkSize) + chunkSize) {
                logger.error("The chunks in the POST is the last one, and the fil is not the correct size")
                return false
            }
            // The file is only a single chunk, and the data size does not fit
            if (numberOfChunks == 1 && file.size != totalSize) {
                logger.error("The file is only a single chunk, and the data size does not fit")
                return false
            }
        }
        return true
    }
}